"""

Dense artificial neural networks. The following models are implemented at the
moment

- Multiple Layer Perceptron  - `DenseNNet`
- Denoising Autoencoder  - `DenoisingAutoencoder`
- stacked Denoising Autoencoder - `StackedDenoisingAutoencoder`

"""

from collections.abc import Sequence, Callable
from functools import reduce
import itertools
from itertools import filterfalse

import lasagne
import numpy as np
import theano

from nnet.base import BaseNNet
from nnet.util import yield_batches, tensor_from_array


__all__ = ["DenseNNet", "DenoisingAutoencoder", "StackedDenoisingAutoencoder"]


class DenseNNet(BaseNNet):
    def __init__(self, input_shape, n_hidden, n_visible,
                 hid_nonlin=lasagne.nonlinearities.rectify,
                 vis_nonlin=lasagne.nonlinearities.softmax,
                 w_init=lasagne.init.GlorotUniform(),
                 dropout=None):
        """
        This is a deep generalisation of the multilayer perceptron model (MLP).
        It allows and arbitrary number of hidden layers of arbitrary size,
        any kind of nonlinear activations in the hidden and visible layers, as
        well as any sort of weight initialisation. And, finally, it supports
        dropout.

        :type input_shape: Sequence[Optional[int]]
        :param input_shape: Input shape. Should be a sequence (a tuple or a
                            list) of positive integers. The very first item in
                            the sequence, which is always interpreted as
                            the batchsize, is the only one that can be set
                            to `None`. That means that the user can't specify
                            any particular batchsize at the initialisation time,
                            though this information allows `Theano` to do some
                            optimisations.
        :type n_hidden: Union[int, Sequence[int]]
        :param n_hidden: the number of units in each hidden layer; it can be a
                         single positive integer or a sequence of positive
                         integer (each integer specifies the number of nodes
                         in the corresponding hidden layer).
        :type n_visible: int
        :param n_visible: the number of visible units
        :type hid_nonlin: Callable
        :param hid_nonlin: any nonliniearity from `lasagne.nonlinearities`
        :type w_init: Union[Callable, Sequence[Callable]]
        :param w_init: A weights initialiser or a Sequence of initialisers;
                       in the first case it's used for all hidden and visible
                       layers; in the later case there should
                       `len(n_hidden) + 1` individual initialisers (one per each
                       hidden layer and one for the visible layer).
        :type dropout: Optional[Union[float, Sequence[float]]
        :param dropout: the dropout probability; an optional real number in
                        [0, 1] or a Sequence of such.
        """

        if not self._is_valid_input_shape(input_shape):
            raise ValueError("Wrong `input_shape`")

        if isinstance(n_hidden, int) and n_hidden <= 0:
            raise ValueError("`n_hidden` must be a either a positive integer"
                             "or a Sequence of such")
        if (isinstance(n_hidden, Sequence) and
                (not n_hidden or
                 not all(n > 0 and isinstance(n, int) for n in n_hidden))):
            raise ValueError("`n_hidden` must be a either a positive integer"
                             "or a Sequence of such")

        if not isinstance(n_visible, int) or n_visible <= 0:
            raise ValueError("`n_visible` must be a positive integer")

        if not all(isinstance(fn, Callable) for fn in (hid_nonlin, vis_nonlin)):
            raise ValueError("`hid_nonlin` and `vis_nonlin` must be callable")

        # TODO make validation methods to allow for flexible inheritance
        # TODO add raises fields to the docs

        # note: all layer variables start with l_
        # note: _n_hidden is always a Sequence of ints (layer sizes)
        self._input_shape = input_shape
        self._batch = input_shape[0]
        self._n_hidden = ([n_hidden] if isinstance(n_hidden, int) else
                          list(n_hidden))
        self._n_visible = n_visible
        self._hid_nonlin = hid_nonlin
        self._vis_nonlin = vis_nonlin
        self._w_init = w_init

        if not self._is_valid_dropout(dropout):
            raise ValueError("`dropout` must be an optional float in [0, 1] "
                             "or a Sequence of such floats of length "
                             "n = 1 + hidden layers count")
        self._dropout = (dropout if isinstance(dropout, Sequence) else
                         [dropout] * (len(self._n_hidden) + 1))

        # if w_init is a Sequence of individual initialisers, check whether
        # their count matches the number of hidden layers + 1 (for
        # the output layer)
        if (isinstance(w_init, Sequence) and
                len(w_init) != len(self._n_hidden) + 1):

            raise ValueError("The number of individual initialisers != "
                             "the number of hidden layers + 1")

        self._network = self._init()

    @staticmethod
    def _is_valid_dropout(dropout):
        def valid_element(x):
            return (x is None) or (isinstance(x, float) and (0.0 <= x <= 1.0))

        return (valid_element(dropout) if not isinstance(dropout, Sequence)
                else all(map(valid_element, dropout)))

    @staticmethod
    def _is_valid_input_shape(input_shape):
        if not isinstance(input_shape, Sequence):
            return False
        elif (input_shape[0] is not None and
                  (not isinstance(input_shape[0], int) or input_shape[0] <= 0)):
            return False
        elif any(not isinstance(val, int) for val in input_shape[1:]):
            return False
        elif any(val <= 0 for val in input_shape[1:]):
            return False
        return True

    def _init(self):
        """
        Initialise the network. This method take a `BaseNNet` instance and
        returns a `lasagne` layer correpsonding to the output layer of the
        network. This method is overloaded in subclasses to create different
        models.

        :rtype: lasagne.layers.DenseLayer
        """
        # note: self._dropout is a list of Optional[float],
        #       self._dropout[0] – the dropout level on the input layer
        #       self._dropout[1:] – dropout levels of the hidden layers

        # create an input layer
        l_inp_raw = lasagne.layers.InputLayer(shape=self._input_shape)
        l_input = (
            lasagne.layers.DropoutLayer(l_inp_raw, p=self._dropout[0]) if
            self._dropout[0] else l_inp_raw)

        # create the hidden facade
        # note: if `hidden` is a Sequence if ints, a facade of `len(hidden)` is
        #       initialised, hence `self._l_hidden_facade` points to the top
        #       (last) layer of the facade.
        hid_w_init = self._make_hid_w_init(self._w_init)
        l_hidden_facade = self._stack_hidden_layers(l_input, self._n_hidden,
                                                    self._hid_nonlin,
                                                    hid_w_init,
                                                    self._dropout[1:])
        # create the visible (output) layer
        vis_w_init = self._make_vis_w_init(self._w_init)
        l_visible = lasagne.layers.DenseLayer(l_hidden_facade,
                                              num_units=self._n_visible,
                                              nonlinearity=self._vis_nonlin,
                                              W=vis_w_init)
        # note: since in `lasagne` it's sufficient to have the output layer
        #       to access the entire network, we add a link-attribute
        #       `network = self._output_shape` to make our intention to refer to
        #       the entire network explicit, when needed.
        return l_visible

    def _make_hid_w_init(self, w_init):
        """
        :type w_init: Union[Callable, Sequence[Callable]]
        :rtype: tuple[Callable]
        """
        hid_w_init = tuple(itertools.repeat(w_init, len(self._n_hidden)) if
                           not isinstance(w_init, Sequence) else w_init[:-1])
        return hid_w_init

    def _make_vis_w_init(self, w_init):
        """
        :type w_init: Union[Callable, Sequence[Callable]]
        :rtype: Callable
        """
        output_w_init = (w_init if not isinstance(w_init, Sequence)
                         else w_init[-1])
        return output_w_init

    # TODO replace `dump` and `load` with `__getstate__` and `__setstate__`

    # def dump(self, fp, compress=1):
    #     """
    #     :type fp: str
    #     :param fp: a filepath to dump to
    #     :type compress: int
    #     :param compress: compression level (1 to 9)
    #     :return: a list of file names in which the data are stored.
    #     :return
    #     """
    #     if not 1 <= compress <= 9:
    #         raise ValueError("Unsupported compression level")
    #
    #     params = {
    #         "input_shape": self._input_shape,
    #         "n_hidden": self._n_hidden,
    #         "n_visible": self._n_visible,
    #         "dropout": self._dropout,
    #         "hid_nonlin": self._hid_nonlin,
    #         "vis_nonlin": self._vis_nonlin,
    #         "w_init": list(self._hid_w_init) + [self._vis_w_init],
    #         "net_params": lasagne.layers.get_all_param_values(self.network)
    #     }
    #
    #     return joblib.dump(params, fp, compress=compress)
    #
    # @classmethod
    # def load(cls, fp):
    #     """
    #     :type fp: str
    #     :param fp: path to a dumped network
    #     :rtype: DenseNNet
    #     :return: a restored neural network
    #     """
    #     params = joblib.load(fp)
    #     net_params = params["net_params"]
    #     network = DenseNNet(**{key: value for key, value in params.iteritems()
    #                            if key in cls.__init__.__code__.co_varnames})
    #     lasagne.layers.set_all_param_values(network.network, net_params)
    #     return network

    # def __getstate__(self):
    #     raise NotImplementedError
    #
    # def __setstate__(self):
    #     raise NotImplementedError

    def _train(self, epochs, x, y, batchsize, update_fn, loss_fn,
               learning_rate, deterministic=False, **kwargs):
        # TODO docs
        # TODO add validation accuracy
        """
        :param epochs:
        :type train_data: (np.ndarray, np.ndarray)
        :param train_data:
        :type validation_data: (np.ndarray, np.ndarray)
        :param train_data:
        :type crossval: Optional[float]
        :param crossval: the fraction of input data used for cross-validation
        :type acc_fn:
        :param acc_fn: a Theano function
        """

        # create target var
        # note: I use my own `tensor_from_array` instead of `theano.shared`,
        #       because for whatever reason Theano says I can't use a shared
        #       variable here and that I should pass it via the `givens`
        #       parameter, whatever that is.
        input_var = lasagne.layers.get_all_layers(self.network)[0].input_var
        target_var = tensor_from_array(y)

        # note: self.network is a getter that returns the output layer of
        #       the `lasagne`-backed network.

        # training functions
        prediction = lasagne.layers.get_output(self.network,
                                               deterministic=deterministic)
        loss = loss_fn(prediction, target_var).mean()
        params = lasagne.layers.get_all_params(self.network, trainable=True)
        updates = update_fn(loss, params, learning_rate=learning_rate, **kwargs)
        train_fn = theano.function([input_var, target_var],
                                   loss, updates=updates)

        def run_epoch(x_, y_):
            train_batches = yield_batches((x_, y_), batchsize)
            train_err = np.mean([train_fn(*batch) for batch in train_batches])
            return train_err

        return (run_epoch(x, y) for _ in itertools.repeat(None, epochs))

    def train(self, epochs, x, y, batchsize,
              update_fn=lasagne.updates.nesterov_momentum,
              loss_fn=lasagne.objectives.categorical_crossentropy,
              learning_rate=0.1, deterministic=False, **kwargs):
            """
            :type epochs: int
            :param epochs: the number of training iterations
            :type x: np.ndarray
            :param x:
            :type y: np.ndarray
            :param y:
            :param batchsize:
            :param update_fn:
            :param loss_fn:
            :param deterministic: turn off dropout during training
            :param kwargs:
            :return:
            :rtype: Generator[float]
            """
            if not isinstance(epochs, int) or epochs < 0:
                raise ValueError("`epochs` must be a non-negative integer")

            if (self._batch is not None and
                    self._batch != batchsize):
                raise ValueError

            if x.shape[0] != y.shape[0]:
                raise ValueError("`x` and `y` must have the same size along"
                                 " the first dimension")

            return self._train(epochs, x, y, batchsize, update_fn,
                               loss_fn, learning_rate, deterministic, **kwargs)

    def feed(self, input_data, i=None, ignore_dropout=True, deterministic=True):
        """
        :type input_data: np.ndarray
        :param input_data: a numpy array to feed to the network
        :param i: the index of the layer to feed; by default the last (the
                  output) layer is used (dropout-layers are ignored)
        :return: The `input_data`'s representation at the `i`-th layer of the
                 network
        :rtype: np.ndarray
        """

        try:
            out_layer = self.network if i is None else (
                self.layers[i] if ignore_dropout else
                lasagne.layers.get_all_layers(self.network)[i])
        except (IndexError, TypeError, ValueError):
            raise IndexError("Index {} is invalid or out of bounds".format(i))

        prediction = lasagne.layers.get_output(out_layer,
                                               deterministic=deterministic)
        prediction_fn = theano.function([self.input_var], prediction)

        if self._batch is not None:
            batches = (b[0] for b in yield_batches((input_data,),
                                                   self._batch, shuffle=False))
            return np.vstack(map(prediction_fn, batches))

        return prediction_fn(input_data)

    @property
    def layers(self):
        all_layers = lasagne.layers.get_all_layers(self.network)
        return list(
            filterfalse(lambda l: isinstance(l, lasagne.layers.DropoutLayer),
                        all_layers)
        )

    @staticmethod
    def _stack_hidden_layers(input_l, units, nonlinearity, initialisers,
                             dropouts):
        """
        Create and stack hidden layers on top of one another.

        :type input_l: lasagne.layers.InputLayer
        :param input_l: the input layer
        :type: units: Sequence[int]
        :param: the number of units per hidden layers
        :type nonlinearity: Callable
        :param: some nonlinearity to use in the hidden layers
        :type initialisers: Sequence[Callable]
        :param initialisers: weight initialisers for each hidden layer
        :type dropouts: Sequence[Optional[float]]
        :param dropouts: the dropout level
        :rtype: lasagne.layers.DenseLayer
        """
        if not len(units) == len(initialisers) == len(dropouts):
            raise ValueError("The numbers of layers, initialisers and dropouts "
                             "do not match")

        def stack_layers(inp_l, n_units_w_init_dropout):
            n_units, w_init_, dropout_ = n_units_w_init_dropout
            layer = lasagne.layers.DenseLayer(inp_l,
                                              num_units=n_units, W=w_init_,
                                              nonlinearity=nonlinearity)
            return (layer if not dropout_ else
                    lasagne.layers.DropoutLayer(layer, p=dropout_))

        return reduce(stack_layers, zip(units, initialisers, dropouts), input_l)

    @property
    def network(self):
        """
        :rtype: lasagne.layers.DenseLayer
        :return: the output layer of the network
        """
        return self._network

    @property
    def input_var(self):
        return lasagne.layers.get_all_layers(self.network)[0].input_var

    @property
    def input_shape(self):
        """
        :rtype: Sequence[Optional[int]]
        """
        return self._input_shape


class DenoisingAutoencoder(DenseNNet):
    def __init__(self, input_shape, n_hidden, dropout=0.5):

        if len(input_shape) != 2:
            raise ValueError("Only 1D inputs are supported at the moment")

        n_visible = input_shape[1]

        # note: noise is only added to the input layer
        super().__init__(input_shape=input_shape, n_hidden=n_hidden,
                         n_visible=n_visible,
                         hid_nonlin=lasagne.nonlinearities.sigmoid,
                         vis_nonlin=lasagne.nonlinearities.sigmoid,
                         w_init=lasagne.init.GlorotUniform(),
                         dropout=[dropout, None])
        pass

    def _init(self):
        if isinstance(self._n_hidden, Sequence) and len(self._n_hidden) != 1:
            raise ValueError(
                "{} instances only support a single hidden "
                "layer due to training efficiency.".format(type(self))
            )

        # create an input layer

        l_inp_raw = lasagne.layers.InputLayer(shape=self._input_shape)
        l_input = (
            lasagne.layers.DropoutLayer(l_inp_raw, p=self._dropout[0]) if
            self._dropout[0] else l_inp_raw)

        # create the hidden facade
        # note: if `hidden` is a Sequence if ints, a facade of `len(hidden)` is
        #       initialised, hence `self._l_hidden_facade` points to the top
        #       (last) layer of the facade.
        hid_w_init = self._make_hid_w_init(self._w_init)
        l_hidden_facade = self._stack_hidden_layers(l_input, self._n_hidden,
                                                    self._hid_nonlin,
                                                    hid_w_init,
                                                    self._dropout[1:])

        # create the visible (output) layer
        vis_w_init = self._make_vis_w_init(l_hidden_facade)
        l_visible = lasagne.layers.DenseLayer(l_hidden_facade,
                                              num_units=self._n_visible,
                                              nonlinearity=self._vis_nonlin,
                                              W=vis_w_init)
        # note: since in `lasagne` it's sufficient to have the output layer
        #       to access the entire network, we add a link-attribute
        #       `network = self._output_shape` to make our intention to refer to
        #       the entire network explicit, when needed.
        return l_visible

    def _make_vis_w_init(self, l_hidden_facade):
        return l_hidden_facade.W.T

    def train(self, epochs, x, batchsize,
              update_fn=lasagne.updates.nesterov_momentum,
              loss_fn=lasagne.objectives.squared_error,
              learning_rate=0.1, **kwargs):
        """
        :type epochs: int
        :type batchsize: int
        :type update_fn: Callable
        :type loss_fn: Callable
        :type learning_rate: float
        :rtype: Generator[float]
        """
        return super().train(epochs=epochs, x=x, y=x, batchsize=batchsize,
                             update_fn=update_fn, loss_fn=loss_fn,
                             learning_rate=learning_rate, deterministic=True)

    @property
    def hidden_layer(self):
        return self.layers[1]


class StackedDenoisingAutoencoder(DenoisingAutoencoder):
    def __init__(self, input_shape, n_hidden, dropout=0.5):
        """
        :type input_shape: tuple[Optional[int]]
        :type n_hidden: Union[int, Sequence[int]]
        """
        super().__init__(input_shape, n_hidden, dropout=dropout)

    def _init(self, *args, **kwargs):
        # note: in a StackedAutoencoder the _network instance attribute is empty
        #       before training
        return None

    def _init_autoencoders(self):
        n_input = (self.input_shape[1], ) + tuple(self._n_hidden[:-1])
        input_shapes = [(self._batch, n_inp) for n_inp in n_input]
        n_hidden = self._n_hidden
        noise = self._dropout[0]

        return [DenoisingAutoencoder(inp_shape, n_hid, noise)
                for inp_shape, n_hid in zip(input_shapes, n_hidden)]

    def _pretrain(self, autoencoders, x, epochs, batchsize, update_fn, loss_fn,
                  learning_rate, **kwargs):

        def work(latent_x, dae):
            """
            :type dae: DenoisingAutoencoder
            :type latent_x: np.ndarray
            """
            errors = list(dae.train(epochs, latent_x, batchsize, update_fn,
                                    loss_fn, learning_rate, **kwargs))
            return dae.feed(latent_x, 1)

        return reduce(work, autoencoders, x)

    def _stack_autoencoders(self, autoencoders):
        """
        :type autoencoders: Sequence[DenoisingAutoencoder]
        :rtype: lasagne.layers.DenseLayer
        """
        def stack_layers(net, layer):
            """
            :rtype: lasagne.layers.DenseLayer
            """
            layer.input_layer = net
            return layer

        hidden_layers = [dae.hidden_layer for dae in autoencoders]
        stacked_layers = reduce(stack_layers, hidden_layers)
        l_vis = lasagne.layers.DenseLayer(stacked_layers, self.input_shape[1],
                                          W=self._w_init,
                                          nonlinearity=self._vis_nonlin)
        return l_vis

    def train(self, epochs, x, batchsize,
              update_fn=lasagne.updates.nesterov_momentum,
              loss_fn=lasagne.objectives.squared_error,
              learning_rate=0.1,  **kwargs):
        """
        Note: you can pass `pretrain=False` to turn pretraining off.
        :type epochs: (int, int)
        :param epochs: pretraining epochs, fine-tuning epochs
        """

        pretrain = kwargs.get("pretrain", True)

        if pretrain not in (True, False):
            raise ValueError("`pretrain` is a boolean flag")

        f_kwargs = {key: value for key, value in kwargs.items()
                    if key != "pretrain"}

        pretrain_epochs, train_epochs = epochs
        autoencoders = self._init_autoencoders()

        if pretrain:
            self._pretrain(autoencoders, x, pretrain_epochs, batchsize,
                           update_fn, loss_fn, learning_rate, **f_kwargs)

        self._network = self._stack_autoencoders(autoencoders)

        return super().train(train_epochs, x, batchsize, update_fn, loss_fn,
                             learning_rate, **f_kwargs)

    @property
    def network(self):
        if self._network is None:
            raise AttributeError("Untrained `StackedAutoencoder` instances have"
                                 "empty `network` attribute before training")
        return self._network


if __name__ == "__main__":
    raise RuntimeError
